package main

import (
	"bufio"
	"fmt"
	"os"
	"strconv"
	"strings"
	"time"

	"github.com/shirou/gopsutil/mem"
	"gopkg.in/ini.v1"
)

var configFilePath = "./config.ini"

type CacheConfig struct {
	Duration      time.Duration
	MaxUsageBytes uint64 // Store as bytes for uniformity
	Path          string
}

type MetaSearchConfig struct {
	Text  []string
	Image []string
	Files []string
	Video []string
}

type Config struct {
	Port                int
	NodeID              string
	Nodes               []string
	Domain              string
	NodesEnabled        bool
	MetaSearchEnabled   bool
	IndexerEnabled      bool
	WebsiteEnabled      bool
	RamCacheEnabled     bool
	DriveCacheEnabled   bool
	MetaProxyEnabled    bool
	MetaProxyStrict     bool
	MetaProxyRetry      int
	MetaProxies         []string
	CrawlerProxyEnabled bool
	CrawlerProxyStrict  bool
	CrawlerProxies      []string
	CrawlerProxyRetry   int
	// Maybye add Proxy support for Image Extraction?
	LogLevel                   int
	ConcurrentStandardCrawlers int
	ConcurrentChromeCrawlers   int
	CrawlingInterval           time.Duration // Refres crawled results in...
	MaxPagesPerDomain          int           // Max pages to crawl per domain
	IndexBatchSize             int
	LibreXInstances            []string

	MetaSearch MetaSearchConfig

	DriveCache CacheConfig
	RamCache   CacheConfig
}

var defaultConfig = Config{
	Port:                       5000,
	Domain:                     "localhost",
	Nodes:                      []string{},
	NodesEnabled:               false,
	MetaSearchEnabled:          true,
	IndexerEnabled:             false,
	WebsiteEnabled:             true,
	RamCacheEnabled:            true,
	DriveCacheEnabled:          false,
	MetaProxyEnabled:           false,
	MetaProxyStrict:            true,
	MetaProxies:                []string{},
	MetaProxyRetry:             3,
	CrawlerProxyEnabled:        false,
	CrawlerProxyStrict:         true,
	CrawlerProxies:             []string{},
	CrawlerProxyRetry:          1,
	ConcurrentStandardCrawlers: 12,
	ConcurrentChromeCrawlers:   4,
	CrawlingInterval:           24 * time.Hour,
	MaxPagesPerDomain:          10,
	IndexBatchSize:             50,
	LogLevel:                   1,
	LibreXInstances:            []string{"librex.antopie.org"},
	MetaSearch: MetaSearchConfig{
		// For Text search (skip SearXNG and LibreX by default, as that would be mega stupid)
		Text: []string{"Google", "Brave", "DuckDuckGo"},

		// For Image search
		Image: []string{"Qwant", "Bing", "DeviantArt"},

		// For Files search
		Files: []string{"TorrentGalaxy", "ThePirateBay"},

		// For Video (piped instances)
		Video: []string{
			"api.piped.yt",
			"pipedapi.moomoo.me",
			"pipedapi.darkness.services",
			"pipedapi.kavin.rocks",
			"piped-api.hostux.net",
			"pipedapi.syncpundit.io",
			"piped-api.cfe.re",
			"pipedapi.in.projectsegfau.lt",
			"piapi.ggtyler.dev",
			"piped-api.codespace.cz",
			"pipedapi.coldforge.xyz",
			"pipedapi.osphost.fi",
		},
	},
	DriveCache: CacheConfig{
		Duration:      48 * time.Hour,                                     // Added
		Path:          "./cache",                                          // Added
		MaxUsageBytes: parseMaxUsageDrive("90 %", config.DriveCache.Path), // Added
	},
	RamCache: CacheConfig{
		Duration:      6 * time.Hour,           // Added
		MaxUsageBytes: parseMaxUsageRam("90%"), // Added
	},
}

func initConfig() error {
	// Check if the configuration file exists
	if _, err := os.Stat(configFilePath); os.IsNotExist(err) {
		// If not, create a new configuration
		return createConfig()
	}

	printInfo("Configuration file already exists. Loading configuration.")
	// Load existing configuration
	config = loadConfig()
	return nil
}

func createConfig() error {
	reader := bufio.NewReader(os.Stdin)

	printMessage("Configuration file not found. Let's set it up.")
	printMessage("Do you want to use default values? (yes/NO): ")
	useDefaults, _ := reader.ReadString('\n')

	if strings.TrimSpace(strings.ToLower(useDefaults)) != "yes" {
		// Server settings
		printMessage("Enter port (default 5000): ")
		portStr, _ := reader.ReadString('\n')
		portStr = strings.TrimSpace(portStr)
		if portStr != "" {
			port, err := strconv.Atoi(portStr)
			if err == nil {
				config.Port = port
			} else {
				printWarn("Invalid port, using default (5000).")
				config.Port = defaultConfig.Port
			}
		} else {
			config.Port = defaultConfig.Port
		}

		printMessage("Enter your domain address (default localhost): ")
		domain, _ := reader.ReadString('\n')
		config.Domain = strings.TrimSpace(domain)
		if config.Domain == "" {
			config.Domain = defaultConfig.Domain
		}

		// printMessage("Use Indexer? (YES/no): ")
		// indexerChoice, _ := reader.ReadString('\n')
		// indexerChoice = strings.TrimSpace(strings.ToLower(indexerChoice))
		// if indexerChoice == "no" {
		// 	config.IndexerEnabled = false
		// } else {
		// 	config.IndexerEnabled = true
		// }

		// Cache settings
		printMessage("Would you like to configure Cache settings (yes/NO): ")
		configureCache, _ := reader.ReadString('\n')
		if strings.TrimSpace(strings.ToLower(configureCache)) == "yes" {
			// RamCache settings
			printMessage("Enter duration to store results in Ram (default 6h): ")
			ramDurationStr, _ := reader.ReadString('\n')
			ramDurationStr = strings.TrimSpace(ramDurationStr)
			if ramDurationStr == "" {
				config.RamCache.Duration = defaultConfig.RamCache.Duration
				config.RamCacheEnabled = defaultConfig.RamCacheEnabled
			} else {
				ramDuration, err := time.ParseDuration(ramDurationStr)
				if err != nil || ramDuration == 0 {
					printWarn("Invalid duration, using default (6h).")
					config.RamCache.Duration = defaultConfig.RamCache.Duration
					config.RamCacheEnabled = defaultConfig.RamCacheEnabled
				} else {
					config.RamCache.Duration = ramDuration
					config.RamCacheEnabled = true
				}
			}

			printMessage("Enter RamCache max usage, e.g., 2 GiB or 80%% (default 90%%): ")
			ramMaxUsage, _ := reader.ReadString('\n')
			ramMaxUsage = strings.TrimSpace(ramMaxUsage)
			if ramMaxUsage == "" {
				config.RamCache.MaxUsageBytes = defaultConfig.RamCache.MaxUsageBytes
			} else if ramMaxUsage == "0" || parseMaxUsageRam(ramMaxUsage) == 0 {
				config.RamCacheEnabled = false
			} else {
				config.RamCache.MaxUsageBytes = parseMaxUsageRam(ramMaxUsage)
				if config.RamCache.MaxUsageBytes == 0 {
					printWarn("Invalid RamCache max usage, using default (90%%).")
					config.RamCache.MaxUsageBytes = defaultConfig.RamCache.MaxUsageBytes
				}
			}

			// DriveCache settings
			printMessage("Enter duration to store results in DriveCache (default 0h): ")
			driveDurationStr, _ := reader.ReadString('\n')
			driveDurationStr = strings.TrimSpace(driveDurationStr)
			if driveDurationStr == "" {
				config.DriveCache.Duration = defaultConfig.DriveCache.Duration
				config.DriveCacheEnabled = defaultConfig.DriveCacheEnabled
			} else {
				driveDuration, err := time.ParseDuration(driveDurationStr)
				if err != nil || driveDuration == 0 {
					printErr("Invalid duration, using default (48h).")
					config.DriveCache.Duration = defaultConfig.DriveCache.Duration
					config.DriveCacheEnabled = defaultConfig.DriveCacheEnabled
				} else {
					config.DriveCache.Duration = driveDuration
					config.DriveCacheEnabled = true
				}
			}

			printMessage("Enter DriveCache path (default ./cache): ")
			drivePath, _ := reader.ReadString('\n')
			drivePath = strings.TrimSpace(drivePath)
			if drivePath == "" {
				config.DriveCache.Path = defaultConfig.DriveCache.Path
			} else {
				config.DriveCache.Path = drivePath
			}

			printMessage("Enter DriveCache max usage, e.g., 2 GiB or 90%% (default 90%%): ")
			driveMaxUsage, _ := reader.ReadString('\n')
			driveMaxUsage = strings.TrimSpace(driveMaxUsage)
			if driveMaxUsage == "" {
				config.DriveCache.MaxUsageBytes = defaultConfig.DriveCache.MaxUsageBytes
			} else if driveMaxUsage == "0" || parseMaxUsageDrive(driveMaxUsage, drivePath) == 0 {
				config.DriveCacheEnabled = false
			} else {
				config.DriveCache.MaxUsageBytes = parseMaxUsageDrive(driveMaxUsage, drivePath)
				if config.DriveCache.MaxUsageBytes == 0 {
					printWarn("Invalid DriveCache max usage, using default.")
					config.DriveCache.MaxUsageBytes = defaultConfig.DriveCache.MaxUsageBytes
				}
			}
		} else {
			printInfo("Cache settings skipped. Using default values.")
			config.RamCache = defaultConfig.RamCache
			config.DriveCache = defaultConfig.DriveCache
		}
	} else {
		// Use default configuration
		config = defaultConfig
	}

	saveConfig(config)
	printInfo("Configuration saved successfully.")
	return nil
}

func saveConfig(config Config) {
	cfg := ini.Empty()

	// Server section
	sec := cfg.Section("Server")
	sec.Key("Port").SetValue(strconv.Itoa(config.Port))
	sec.Key("Domain").SetValue(config.Domain)
	sec.Key("LogLevel").SetValue(strconv.Itoa(config.LogLevel))

	// Nodes section
	nodesSec := cfg.Section("Nodes")
	nodesSec.Key("NodeID").SetValue(config.NodeID)
	nodesSec.Key("Nodes").SetValue(strings.Join(config.Nodes, ","))

	// Features section
	if config.NodesEnabled != defaultConfig.NodesEnabled ||
		config.MetaSearchEnabled != defaultConfig.MetaSearchEnabled ||
		config.IndexerEnabled != defaultConfig.IndexerEnabled ||
		config.WebsiteEnabled != defaultConfig.WebsiteEnabled ||
		config.MetaProxyEnabled != defaultConfig.MetaProxyEnabled ||
		config.CrawlerProxyEnabled != defaultConfig.CrawlerProxyEnabled {

		featuresSec := cfg.Section("Features")
		setBoolIfChanged(featuresSec, "Nodes", config.NodesEnabled, defaultConfig.NodesEnabled)
		setBoolIfChanged(featuresSec, "Crawler", config.MetaSearchEnabled, defaultConfig.MetaSearchEnabled)
		setBoolIfChanged(featuresSec, "Indexer", config.IndexerEnabled, defaultConfig.IndexerEnabled)
		setBoolIfChanged(featuresSec, "Website", config.WebsiteEnabled, defaultConfig.WebsiteEnabled)
		setBoolIfChanged(featuresSec, "MetaProxy", config.MetaProxyEnabled, defaultConfig.MetaProxyEnabled)
		setBoolIfChanged(featuresSec, "CrawlerProxy", config.CrawlerProxyEnabled, defaultConfig.CrawlerProxyEnabled)
	}

	// Proxies section
	proxiesSec := cfg.Section("Proxies")
	proxiesSec.Key("MetaProxyStrict").SetValue(strconv.FormatBool(config.MetaProxyStrict))
	proxiesSec.Key("MetaProxies").SetValue(strings.Join(config.MetaProxies, ","))
	setBoolIfChanged(proxiesSec, "CrawlerProxyStrict", config.CrawlerProxyStrict, defaultConfig.CrawlerProxyStrict)
	setSliceIfChanged(proxiesSec, "CrawlerProxies", config.CrawlerProxies, defaultConfig.CrawlerProxies)
	proxiesSec.Key("MetaProxyRetry").SetValue(strconv.Itoa(config.MetaProxyRetry))
	setIntIfChanged(proxiesSec, "CrawlerProxyRetry", config.CrawlerProxyRetry, defaultConfig.CrawlerProxyRetry)

	// MetaSearch section
	metaSec := cfg.Section("MetaSearches")
	metaSec.Key("LibreXInstances").SetValue(strings.Join(config.LibreXInstances, ","))
	metaSec.Key("Text").SetValue(strings.Join(config.MetaSearch.Text, ","))
	metaSec.Key("Image").SetValue(strings.Join(config.MetaSearch.Image, ","))
	metaSec.Key("Files").SetValue(strings.Join(config.MetaSearch.Files, ","))
	metaSec.Key("Video").SetValue(strings.Join(config.MetaSearch.Video, ","))

	// Indexer section
	if config.ConcurrentStandardCrawlers != defaultConfig.ConcurrentStandardCrawlers ||
		config.ConcurrentChromeCrawlers != defaultConfig.ConcurrentChromeCrawlers ||
		config.CrawlingInterval != defaultConfig.CrawlingInterval ||
		config.MaxPagesPerDomain != defaultConfig.MaxPagesPerDomain ||
		config.IndexBatchSize != defaultConfig.IndexBatchSize {

		indexerSec := cfg.Section("Indexer")
		setIntIfChanged(indexerSec, "ConcurrentStandardCrawlers", config.ConcurrentStandardCrawlers, defaultConfig.ConcurrentStandardCrawlers)
		setIntIfChanged(indexerSec, "ConcurrentChromeCrawlers", config.ConcurrentChromeCrawlers, defaultConfig.ConcurrentChromeCrawlers)
		setIfChanged(indexerSec, "CrawlingInterval", config.CrawlingInterval.String(), defaultConfig.CrawlingInterval.String())
		setIntIfChanged(indexerSec, "MaxPagesPerDomain", config.MaxPagesPerDomain, defaultConfig.MaxPagesPerDomain)
		setIntIfChanged(indexerSec, "IndexBatchSize", config.IndexBatchSize, defaultConfig.IndexBatchSize)
	}

	// DriveCache section
	driveSec := cfg.Section("DriveCache")
	driveSec.Key("Duration").SetValue(config.DriveCache.Duration.String())
	driveSec.Key("MaxUsage").SetValue(formatMaxUsage(config.DriveCache.MaxUsageBytes))
	driveSec.Key("Path").SetValue(config.DriveCache.Path)
	// driveSec.Key("MaxConcurrentDownloads.Thumbnail").SetValue(strconv.Itoa(config.DriveCache.MaxConcurrentThumbnailDownloads))

	// RamCache section
	ramSec := cfg.Section("RamCache")
	ramSec.Key("Duration").SetValue(config.RamCache.Duration.String())
	ramSec.Key("MaxUsage").SetValue(formatMaxUsage(config.RamCache.MaxUsageBytes))

	err := cfg.SaveTo(configFilePath)
	if err != nil {
		printErr("Error writing to config file: %v", err)
	}
}

func loadConfig() Config {
	cfg, err := ini.Load(configFilePath)
	if err != nil {
		printErr("Error opening config file: %v", err)
	}

	// Server
	port := getConfigValue(cfg.Section("Server").Key("Port"), defaultConfig.Port, strconv.Atoi)
	domain := getConfigValueString(cfg.Section("Server").Key("Domain"), defaultConfig.Domain)
	logLevel := getConfigValue(cfg.Section("Server").Key("LogLevel"), defaultConfig.LogLevel, strconv.Atoi)

	// Nodes
	nodeID := getConfigValueString(cfg.Section("Nodes").Key("NodeID"), defaultConfig.NodeID)
	nodes := strings.Split(getConfigValueString(cfg.Section("Nodes").Key("Nodes"), ""), ",")

	// Features
	nodesEnabled := getConfigValueBool(cfg.Section("Features").Key("Nodes"), defaultConfig.NodesEnabled)
	metaSearchEnabled := getConfigValueBool(cfg.Section("Features").Key("Crawler"), defaultConfig.MetaSearchEnabled)
	indexerEnabled := getConfigValueBool(cfg.Section("Features").Key("Indexer"), defaultConfig.IndexerEnabled)
	websiteEnabled := getConfigValueBool(cfg.Section("Features").Key("Website"), defaultConfig.WebsiteEnabled)
	ramCacheEnabled := getConfigValueBool(cfg.Section("Features").Key("RamCache"), defaultConfig.RamCacheEnabled)
	driveCacheEnabled := getConfigValueBool(cfg.Section("Features").Key("DriveCache"), defaultConfig.DriveCacheEnabled)
	metaProxyEnabled := getConfigValueBool(cfg.Section("Features").Key("MetaProxy"), defaultConfig.MetaProxyEnabled)
	crawlerProxyEnabled := getConfigValueBool(cfg.Section("Features").Key("CrawlerProxy"), defaultConfig.CrawlerProxyEnabled)

	// Nodes
	metaProxyStrict := getConfigValueBool(cfg.Section("Proxies").Key("MetaProxyStrict"), defaultConfig.MetaProxyStrict)
	metaProxies := strings.Split(getConfigValueString(cfg.Section("Proxies").Key("MetaProxies"), ""), ",")
	crawlerProxyStrict := getConfigValueBool(cfg.Section("Proxies").Key("CrawlerProxyStrict"), defaultConfig.CrawlerProxyStrict)
	crawlerProxies := strings.Split(getConfigValueString(cfg.Section("Proxies").Key("CrawlerProxies"), ""), ",")
	metaProxyRetry := getConfigValue(cfg.Section("Proxies").Key("MetaProxyRetry"), defaultConfig.MetaProxyRetry, strconv.Atoi)
	crawlerProxyRetry := getConfigValue(cfg.Section("Proxies").Key("CrawlerProxyRetry"), defaultConfig.CrawlerProxyRetry, strconv.Atoi)

	// MetaSearch
	searchXInstances := strings.Split(getConfigValueString(cfg.Section("MetaSearches").Key("LibreXInstances"), strings.Join(defaultConfig.LibreXInstances, ",")), ",")
	textList := strings.Split(getConfigValueString(cfg.Section("MetaSearch").Key("Text"), strings.Join(defaultConfig.MetaSearch.Text, ",")), ",")
	imageList := strings.Split(getConfigValueString(cfg.Section("MetaSearch").Key("Image"), strings.Join(defaultConfig.MetaSearch.Image, ",")), ",")
	filesList := strings.Split(getConfigValueString(cfg.Section("MetaSearch").Key("Files"), strings.Join(defaultConfig.MetaSearch.Files, ",")), ",")
	videoList := strings.Split(getConfigValueString(cfg.Section("MetaSearch").Key("Video"), strings.Join(defaultConfig.MetaSearch.Video, ",")), ",")

	// Load default values for MetaSearch if they are empty
	if len(textList) == 1 && textList[0] == "" {
		textList = defaultConfig.MetaSearch.Text
	}
	if len(imageList) == 1 && imageList[0] == "" {
		imageList = defaultConfig.MetaSearch.Image
	}
	if len(filesList) == 1 && filesList[0] == "" {
		filesList = defaultConfig.MetaSearch.Files
	}
	if len(videoList) == 1 && videoList[0] == "" {
		videoList = defaultConfig.MetaSearch.Video
	}

	// Indexing
	concurrentStandardCrawlers := getConfigValue(cfg.Section("Indexer").Key("ConcurrentStandardCrawlers"), defaultConfig.ConcurrentStandardCrawlers, strconv.Atoi)
	concurrentChromeCrawlers := getConfigValue(cfg.Section("Indexer").Key("ConcurrentChromeCrawlers"), defaultConfig.ConcurrentChromeCrawlers, strconv.Atoi)
	crawlingInterval := getConfigValue(cfg.Section("Indexer").Key("CrawlingInterval"), defaultConfig.CrawlingInterval, time.ParseDuration)
	maxPagesPerDomain := getConfigValue(cfg.Section("Indexer").Key("MaxPagesPerDomain"), defaultConfig.MaxPagesPerDomain, strconv.Atoi)
	indexBatchSize := getConfigValue(cfg.Section("Indexer").Key("IndexBatchSize"), defaultConfig.IndexBatchSize, strconv.Atoi)

	// DriveCache
	driveDuration := getConfigValue(cfg.Section("DriveCache").Key("Duration"), defaultConfig.DriveCache.Duration, time.ParseDuration)
	drivePath := getConfigValueString(cfg.Section("DriveCache").Key("Path"), defaultConfig.DriveCache.Path)
	driveMaxUsage := parseMaxUsageDrive(getConfigValueString(cfg.Section("DriveCache").Key("MaxUsage"), formatMaxUsage(defaultConfig.DriveCache.MaxUsageBytes)), drivePath)
	// maxConcurrentDownloads, _ := cfg.Section("DriveCache").Key("MaxConcurrentDownloads.Thumbnail").Int()
	// if maxConcurrentDownloads == 0 {
	// 	maxConcurrentDownloads = defaultConfig.DriveCache.MaxConcurrentThumbnailDownloads
	// }

	// RamCache
	ramDuration := getConfigValue(cfg.Section("RamCache").Key("Duration"), defaultConfig.RamCache.Duration, time.ParseDuration)
	ramMaxUsage := parseMaxUsageRam(getConfigValueString(cfg.Section("RamCache").Key("MaxUsage"), formatMaxUsage(defaultConfig.RamCache.MaxUsageBytes)))

	return Config{
		Port:                       port,
		Domain:                     domain,
		LogLevel:                   logLevel,
		NodeID:                     nodeID,
		Nodes:                      nodes,
		NodesEnabled:               nodesEnabled,
		MetaSearchEnabled:          metaSearchEnabled,
		IndexerEnabled:             indexerEnabled,
		WebsiteEnabled:             websiteEnabled,
		RamCacheEnabled:            ramCacheEnabled,
		DriveCacheEnabled:          driveCacheEnabled,
		MetaProxyEnabled:           metaProxyEnabled,
		MetaProxyStrict:            metaProxyStrict,
		MetaProxies:                metaProxies,
		MetaProxyRetry:             metaProxyRetry,
		CrawlerProxyEnabled:        crawlerProxyEnabled,
		CrawlerProxyStrict:         crawlerProxyStrict,
		CrawlerProxies:             crawlerProxies,
		CrawlerProxyRetry:          crawlerProxyRetry,
		ConcurrentStandardCrawlers: concurrentStandardCrawlers,
		ConcurrentChromeCrawlers:   concurrentChromeCrawlers,
		CrawlingInterval:           crawlingInterval,
		MaxPagesPerDomain:          maxPagesPerDomain,
		IndexBatchSize:             indexBatchSize,
		LibreXInstances:            searchXInstances,
		MetaSearch: MetaSearchConfig{
			Text:  textList,
			Image: imageList,
			Files: filesList,
			Video: videoList,
		},
		DriveCache: CacheConfig{
			Duration:      driveDuration,
			MaxUsageBytes: driveMaxUsage,
			Path:          drivePath,
		},
		RamCache: CacheConfig{
			Duration:      ramDuration,
			MaxUsageBytes: ramMaxUsage,
		},
	}
}

func setIfChanged(sec *ini.Section, key string, value string, defaultValue string) {
	if value != defaultValue {
		sec.Key(key).SetValue(value)
	}
}

func setBoolIfChanged(sec *ini.Section, key string, value bool, defaultValue bool) {
	if value != defaultValue {
		sec.Key(key).SetValue(strconv.FormatBool(value))
	}
}

func setIntIfChanged(sec *ini.Section, key string, value int, defaultValue int) {
	if value != defaultValue {
		sec.Key(key).SetValue(strconv.Itoa(value))
	}
}

func setSliceIfChanged(sec *ini.Section, key string, value, defaultValue []string) {
	if strings.Join(value, ",") != strings.Join(defaultValue, ",") {
		sec.Key(key).SetValue(strings.Join(value, ","))
	}
}

// getConfigValue retrieves a configuration value or returns a default value from defaultConfig.
func getConfigValue[T any](key *ini.Key, defaultValue T, parseFunc func(string) (T, error)) T {
	if key == nil || key.String() == "" {
		return defaultValue
	}
	value, err := parseFunc(key.String())
	if err != nil {
		return defaultValue
	}
	return value
}

// getConfigValueString retrieves a string value or falls back to the default.
func getConfigValueString(key *ini.Key, defaultValue string) string {
	if key == nil || key.String() == "" {
		return defaultValue
	}
	return key.String()
}

// getConfigValueBool retrieves a boolean value or falls back to the default.
func getConfigValueBool(key *ini.Key, defaultValue bool) bool {
	if key == nil || key.String() == "" {
		return defaultValue
	}
	return key.MustBool(defaultValue)
}

// Helper to parse MaxUsage string into bytes
func parseMaxUsageRam(value string) uint64 {
	const GiB = 1024 * 1024 * 1024
	value = strings.TrimSpace(value)
	valueNoSpaces := strings.ReplaceAll(value, " ", "")

	if strings.HasSuffix(valueNoSpaces, "%") {
		percentStr := strings.TrimSuffix(valueNoSpaces, "%")
		percent, err := strconv.ParseFloat(percentStr, 64)
		if err != nil {
			return 0
		}
		totalMem := getTotalMemory()
		return uint64(float64(totalMem) * (percent / 100))
	} else if strings.HasSuffix(valueNoSpaces, "GiB") {
		sizeStr := strings.TrimSuffix(valueNoSpaces, "GiB")
		size, err := strconv.ParseFloat(sizeStr, 64)
		if err != nil {
			return 0
		}
		return uint64(size * GiB)
	}

	return 0
}

// Helper to parse MaxUsage string into bytes based on drive space
func parseMaxUsageDrive(value string, cachePath string) uint64 {
	const GiB = 1024 * 1024 * 1024
	value = strings.TrimSpace(value)
	valueNoSpaces := strings.ReplaceAll(value, " ", "")

	totalDiskSpace := getTotalDiskSpace(cachePath)
	if totalDiskSpace == 0 {
		printErr("Failed to retrieve disk space for path: %s", cachePath)
		return 0
	}

	if strings.HasSuffix(valueNoSpaces, "%") {
		percentStr := strings.TrimSuffix(valueNoSpaces, "%")
		percent, err := strconv.ParseFloat(percentStr, 64)
		if err != nil {
			return 0
		}
		return uint64(float64(totalDiskSpace) * (percent / 100))
	} else if strings.HasSuffix(valueNoSpaces, "GiB") {
		sizeStr := strings.TrimSuffix(valueNoSpaces, "GiB")
		size, err := strconv.ParseFloat(sizeStr, 64)
		if err != nil {
			return 0
		}
		return uint64(size * GiB)
	}

	return 0
}

// Helper to format bytes back to human-readable string
func formatMaxUsage(bytes uint64) string {
	const GiB = 1024 * 1024 * 1024
	if bytes >= GiB {
		return fmt.Sprintf("%.2fGiB", float64(bytes)/GiB)
	}
	return fmt.Sprintf("%dbytes", bytes)
}

// Get total memory of the system
func getTotalMemory() uint64 {
	v, err := mem.VirtualMemory()
	if err != nil {
		printErr("Failed to retrieve system memory: %v", err)
		return 0
	}
	return v.Total
}
